class NonDictInputError(Exception):
    """Исключение - аргумент функции не является словарем"""
    def __str__(self):
        return 'Аргумент функции должен быть словарем'


class IncorrectDataReceivedError(Exception):
    """Исключение - некорректное сообщение"""
    def __str__(self):
        return 'Принято некорректное сообщение от удаленного компьютера'


class ReqFieldMissingError(Exception):
    """Исключение - в запросе отсутствует обязательное поле"""
    def __init__(self, missing_field):
        self.missing_field = missing_field

    def __str__(self):
        return f'В принятом словаре отсутствует обязательное поле {self.missing_field}'

class ServerError(Exception):
    """Исключение - ошибка сервера"""
    def __init__(self, text):
        self.text = text

    def __str__(self):
        return self.text



from sqlalchemy import create_engine, Column, Integer, String, ForeignKey, DateTime, Text, BINARY
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
import datetime

Base = declarative_base()


class ServerStorage:
    """Класс хранилища сервера"""

    class Users(Base):
        """Таблица пользователя"""
        __tablename__ = 'Users'
        id = Column(Integer, primary_key=True)
        login = Column(String, unique=True)
        info = Column(String)
        last_login = Column(DateTime)
        pubkey = Column(Text)
        passwd_hash = Column(String)

        def __init__(self, login, passwd_hash, info=''):
            self.login = login
            self.info = info
            self.last_login = None
            self.pubkey = None
            self.passwd_hash = passwd_hash

        def __repl__(self):
            return f'User{self.id}({self.login})'

    class UserLoginHistory(Base):
        """Таблица истории входов"""
        __tablename__ = 'user_login_history'
        id = Column(Integer, primary_key=True)
        user_id = Column(ForeignKey('Users.id'))
        ipaddress = Column(String)
        port = Column(Integer)
        login_time = Column(DateTime)

        def __init__(self, user_id, ipaddress, port, login_time):
            self.user_id = user_id
            self.ipaddress = ipaddress
            self.port = port
            self.login_time = login_time

        def __repl__(self):
            return f'User{self.user_id}-ip{self.ipaddress}:{self:port}-{self.login_time}'

    class UsersHistory(Base):
        """Таблица счетчик отправленых и полученых сообщений"""
        __tablename__ = 'history'
        id = Column(Integer, primary_key=True)
        user = Column(ForeignKey('Users.id'))
        sent = Column(Integer)
        accepted = Column(Integer)

        def __init__(self, user):
            self.user = user
            self.sent = 0
            self.accepted = 0

    class ActiveUsers(Base):
        """Таблица подключеных пользователей"""
        __tablename__ = 'active_users'
        id = Column(Integer, primary_key=True)
        user_id = Column(ForeignKey('Users.id'), unique=True)
        ipaddress = Column(String)
        port = Column(Integer)
        login_time = Column(DateTime)

        def __init__(self, user_id, ipaddress, port, login_time):
            self.user_id = user_id
            self.ipaddress = ipaddress
            self.port = port
            self.login_time = login_time

        def __repl__(self):
            return f'User{self.user_id}-ip{self.ipaddress}:{self:port}-{self.login_time}'

    class Contacts(Base):
        """Таблица контактов"""
        __tablename__ = 'contacts'
        id = Column(Integer, primary_key=True)
        user_id = Column(ForeignKey('Users.id'))
        contact_id = Column(ForeignKey('Users.id'))

        def __init__(self, user_id, contact_id):
            self.user_id = user_id
            self.contact_id = contact_id

        def __repl__(self):
            return f'User{self.user_id}-contact{self.contact_id}'

    def __init__(self, path):
        self.engine = create_engine(f'sqlite:///{path}',
                                    echo=False,
                                    pool_recycle=7200,
                                    connect_args={'check_same_thread': False}
                                    )
        Base.metadata.create_all(self.engine)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()
        self.session.query(self.ActiveUsers).delete()  # очищаем таблицу активных юзеров при старте
        self.session.commit()

    def user_login(self, username, ipaddress, port, pubkey=None):
        """
        Метод логина пользователя, проверяет что такой пользователь есть (если нет - создает)
        и записывает данные в активные и историю входов
        """
        find_login = self.session.query(self.Users).filter_by(login=username)
        # ищем пользователя с таким же логином
        if find_login.count():  # если есть то берем его логин и записываем в историю входа и в активные
            user = find_login.first()
            user.last_login = datetime.datetime.now()
            # проверяем не изменился ли ключ, если изменился то обновляем
            if user.pubkey != pubkey:
                user.pubkey = pubkey
            new_log_record = self.UserLoginHistory(user.id, ipaddress, port, datetime.datetime.now())
            new_active = self.ActiveUsers(user.id, ipaddress, port, datetime.datetime.now())
            self.session.add(new_log_record)
            self.session.add(new_active)
            self.session.commit()  # все записали
        else:  # если нет то бросаем исключение
            raise ValueError('Пользователь не зарегистрирован.')

    def getactive(self):
        """Метод возвращающий активных пользователей"""
        active_users = self.session.query(self.Users.login, self.ActiveUsers.ipaddress, self.ActiveUsers.port,
                                          self.ActiveUsers.login_time).join(self.Users).all()
        return active_users

    def users_list(self):
        """Метод возвращающий список известных пользователей со временем последнего входа."""
        # Запрос строк таблицы пользователей.
        query = self.session.query(
            self.Users.login,
            self.Users.last_login
        )
        # Возвращаем список кортежей
        return query.all()

    def getall(self):
        """Метод возвращающий всех пользователй"""
        all_users = self.session.query(self.Users.id, self.Users.login).all()
        return all_users

    def history_log(self, name=''):
        """Метод возвращающий историю входов"""
        logs = self.session.query(self.Users.login, self.UserLoginHistory.ipaddress, self.UserLoginHistory.port,
                                  self.UserLoginHistory.login_time).join(self.Users)

        return logs.all() if name == '' else logs.filter_by(login=name).all()

    def add_user(self, name, passwd_hash):
        """
        Метод регистрации пользователя.
        Принимает имя и хэш пароля, создаёт запись в таблице статистики.
        """
        user_row = self.Users(name, passwd_hash)
        self.session.add(user_row)
        self.session.commit()
        history_row = self.UsersHistory(user_row.id)
        self.session.add(history_row)
        self.session.commit()

    def remove_user(self, name):
        """Метод удаляющий пользователя из базы."""
        user = self.session.query(self.Users).filter_by(login=name).first()
        self.session.query(self.ActiveUsers).filter_by(id=user.id).delete()
        self.session.query(self.UserLoginHistory).filter_by(user_id=user.id).delete()
        self.session.query(self.Contacts).filter_by(user_id=user.id).delete()
        self.session.query(
            self.Contacts).filter_by(
            contact_id=user.id).delete()
        self.session.query(self.UsersHistory).filter_by(user=user.id).delete()
        self.session.query(self.Users).filter_by(login=name).delete()
        self.session.commit()

    def get_hash(self, name):
        """Метод получения хэша пароля пользователя."""
        user = self.session.query(self.Users).filter_by(login=name).first()
        return user.passwd_hash

    def get_pubkey(self, name):
        """Метод получения публичного ключа пользователя."""
        user = self.session.query(self.Users).filter_by(login=name).first()
        return user.pubkey

    def check_user(self, name):
        """Метод проверяющий существование пользователя."""
        if self.session.query(self.Users).filter_by(login=name).count():
            return True
        else:
            return False

    def user_logout(self, username):
        """Метод обрабатывающий выход пользователя. Удаляет из таблицы активных"""
        try:
            # Запрашиваем пользователя, что покидает нас
            user = self.session.query(self.Users).filter_by(login=username).first()
            # Удаляем его из таблицы активных пользователей.
            self.session.query(self.ActiveUsers).filter_by(user_id=user.id).delete()
            self.session.commit()
        except:
            pass

    def add_contact(self, user, contact):
        """Метод добавления контакта пользователя"""
        user = self.session.query(self.Users).filter_by(login=user).first()
        contact = self.session.query(self.Users).filter_by(login=contact).first()
        # Проверяем что не дубль и что контакт может существовать (полю пользователь мы доверяем)
        if not contact or self.session.query(self.Contacts).filter_by(user_id=user.id, contact_id=contact.id).count():
            print('Такой контакт уже есть')
            return
        else:
            new_contact = self.Contacts(user.id, contact.id)
            self.session.add(new_contact)
            self.session.commit()

    def del_contact(self, user, contact):
        """Метод удаления контакта пользователя"""
        # Получаем ID пользователей
        user = self.session.query(self.Users).filter_by(login=user).first()
        contact = self.session.query(self.Users).filter_by(login=contact).first()

        # Проверяем что контакт может существовать (полю пользователь мы доверяем)
        if not contact:
            print('Нет такого котнтакта')
            return
        # Удаляем контакт
        print(self.session.query(self.Contacts).
              filter(self.Contacts.user_id == user.id, self.Contacts.contact_id == contact.id).delete())
        self.session.commit()

    def message_history(self):
        """Метод возвращающий данные по отправленным и принятым сообщениям пользователя"""
        query = self.session.query(
            self.Users.login,
            self.Users.last_login,
            self.UsersHistory.sent,
            self.UsersHistory.accepted
        ).join(self.Users, self.Users.id == self.UsersHistory.user)
        # Возвращаем список кортежей
        # print(query.all())
        return query.all()
        # return [('user1', '2022-11-28 01:49:11.843131', 9, 10)]

    def get_user_contacts(self, user):
        """Метод получения контактов пользователя"""
        try:
            # ищем указанного пользователя
            user = self.session.query(self.Users).filter_by(login=user).one()
            # contacts = self.session.query(self.Contacts.user_id,).filter_by(user_id=user_id).all()
        except:
            contacts = []
        else:
            # Запрашиваем его список контактов
            query = self.session.query(self.Contacts.contact_id, self.Users.login). \
                filter_by(user_id=user.id). \
                join(self.Users, self.Contacts.contact_id == self.Users.id)
            # выбираем только имен контактов и возвращаем
            contacts = [contact[1] for contact in query.all()]
            print(contacts)
        return contacts

    def process_message(self, sender, recipient):
        """Метод обновляющий счетчики отправленных и полученых сообщений"""
        # Получаем ID отправителя и получателя
        sender = self.session.query(self.Users).filter_by(login=sender).first().id
        recipient = self.session.query(self.Users).filter_by(login=recipient).first().id
        # Запрашиваем строки из истории и увеличиваем счётчики
        sender_row = self.session.query(self.UsersHistory).filter_by(user=sender).first()
        sender_row.sent += 1
        recipient_row = self.session.query(self.UsersHistory).filter_by(user=recipient).first()
        recipient_row.accepted += 1

        self.session.commit()


if __name__ == '__main__':
    client = ServerStorage('server_base.db3')
    # print(client.getactive())
    # client.user_login('ivanov', '127.0.0.1', 55)
    # client.user_login('pppetrov', '127.0.0.1', 11)
    # client.user_login('kuznetsov', '127.0.0.127', 99)
    # client.user_login('pppetroff', '127.0.0.99', 22)
    # # print(client.getall()))
    # print(client.getactive())
    # client.user_logout(1)
    # print(client.getactive())
    # print(client.history_log())
    # print(client.history_log('ivanov'))
    # print(client.get_user_contacts(1))
    # client.add_contact('ivanov', 'pppetrov')
    # client.add_contact(1, 3)
    # client.add_contact(1, 4)
    # print(client.get_user_contacts(1))
    # client.del_contact(1, 2)
    # client.del_contact(1, 4)
    # client.add_contact(1, 3)
    # client.del_contact(1, 4)
    # print(client.get_user_contacts('ivanov'))
    client.message_history()



import inspect
import logging
import sys
import traceback


if sys.argv[0].find('client.py') == -1:
    LOGGER = logging.getLogger('server')
else:
    LOGGER = logging.getLogger('client')


def func_log(in_function):
    """функция декоратор логирующая вызовы функций"""
    def wrapper(*args, **kwargs):
        result = in_function(*args, **kwargs)
        LOGGER.debug(f"Вызвана функция {in_function.__name__} "
                     f"из функции {traceback.format_stack()[0].strip().split()[-1]} "
                     f"а точнее из {inspect.stack()[1][3]} с параметрами {args} {kwargs},"
                     f" вызов из модуля {in_function.__name__}", stacklevel=2)
        return result
    return wrapper


class Log:
    def __call__(self, in_function):
        """функция декоратор логирующая вызовы методов"""
        def wrapper(*args, **kwargs):
            result = in_function(*args, **kwargs)
            LOGGER.debug(f"Вызвана функция {in_function.__name__} "
                         f"из функции {traceback.format_stack()[0].strip().split()[-1]} "
                         f"а точнее из {inspect.stack()[1][3]} с параметрами {args} {kwargs},"
                         f" вызов из модуля {in_function.__name__}", stacklevel=2)
            return result
        return wrapper



class IsPortValid:
    """Класс-дескриптор, проверяющий, что заданный порт является целым положительным числом"""
    def __get__(self, instance, owner):
        return instance.__dict__[self.my_attr]

    def __set__(self, instance, value):
        try:
            new_value = int(value)
        except ValueError:
            raise ValueError("Порт должен быть числом")
        else:
            if new_value != value:
                # raise ValueError("Порт должен быть ЦЕЛЫМ числом")
                print(f"Вы задали порт:{value}. Порт должен быть ЦЕЛЫМ числом. Будет установлен порт по умолчанию: 7777")
                new_value = 7777
            if value < 0:
                # raise ValueError("Порт не может быть отрицательным")
                print(f"Вы задали порт:{value}. Порт не может быть отрицательным. Будет установлен порт по умолчанию: "
                      f"7777")
                new_value = 7777

            instance.__dict__[self.my_attr] = new_value

    def __set_name__(self, owner, my_attr):
        self.my_attr = my_attr


class Wr:
    port = IsPortValid()

    def __init__(self, port):
        self.port = port

    def __str__(self):
        return f'new port is {self.port}'


if __name__ == '__main__':
    w = Wr(999)
    print(w)







import logging
import os
import select
import socket
import sys
import subprocess
import hmac
import binascii
import threading
import chardet
import configparser

from PyQt5.QtWidgets import QApplication, QMessageBox
from PyQt5.QtCore import QTimer

from .server_gui import MainWindow, gui_create_model, HistoryWindow, create_stat_model, ConfigWindow
dir_path = os.path.dirname(os.path.realpath(__file__))
import_path = os.path.abspath(os.path.join(dir_path, os.pardir))
sys.path.append(import_path)
from common.metaclasses import ServerVerifier
from common.descriptors import IsPortValid
from common.variables import ACTION, ACCOUNT_NAME, MAX_CONNECTIONS, PRESENCE, TIME, USER, ERROR, MESSAGE_TEXT, \
    MESSAGE, SENDER, DESTINATION, RESPONSE_200, RESPONSE_400, EXIT, GETCLIENTS, \
    LIST, RESPONSE_CLIENTS, RESPONSE, GETCONTACTS, RESPONSE_202, ADD_CONTACT, REMOVE_CONTACT, PUBLIC_KEY_REQUEST, DATA, \
    RESPONSE_511, PUBLIC_KEY, RESPONSE_205
from common.utils import get_message, send_message, create_arg_parser
from .add_user import RegisterUser
from .remove_user import DelUserDialog
from .database import ServerStorage

LOGGER = logging.getLogger('server')  # забрали логгер из конфига

# Флаг, что был подключён новый пользователь, нужен чтобы не мучить DB
# постоянными запросами на обновление
new_connection = False
conflag_lock = threading.Lock()


def arg_parser(default_port, default_address):
    LOGGER.info('Разбираем параметры для запуска сервера')
    parser = create_arg_parser(default_port, default_address)
    namespace = parser.parse_args(sys.argv[1:])
    listen_address = namespace.a
    listen_port = namespace.p
    return listen_address, listen_port


def console(server):  # консольный вариант сервера
    help = 'help - справка\nexit - выход\ngetall - список всех пользователей\nconnected - подключенные сейчас\nlog - ' \
           'история сообщений '
    while True:  # поток взаимодействия с оператором сервера
        command = input("help - список всех команд. Введите команду: ")
        if command == 'exit':
            break
        elif command == 'help':
            print(help)
        elif command == 'connected':
            print('Будет реализовано в следующей версии программы')
            pass
        elif command == 'getall':
            for item in server.database.getall():
                print(item)
        elif command == 'log':
            name = input('Введите имя пользователя или нажмите enter если хотите вывести историю всех пользователей: ')
            history_log = server.database.history_log(name)
            for item in history_log:
                print(item)


def main():
    """
    Функция оркестратор запуска сервера. Получает параметры запуска из файла конфигаи
    стратует потоки обработки сообщений и графики
    """
    # Загрузка файла конфигурации сервера
    config = configparser.ConfigParser()
    dir_path = os.path.dirname(os.path.realpath(__file__))
    config.read(f"{dir_path}/{'server.ini'}")

    # Загрузка параметров командной строки, если нет параметров, то задаём
    # значения по умоланию.
    listen_address, listen_port = arg_parser(
        config['SETTINGS']['Default_port'], config['SETTINGS']['Listen_Address'])

    # Инициализация базы данных
    database = ServerStorage(
        os.path.join(
            config['SETTINGS']['Database_path'],
            config['SETTINGS']['Database_file']))

    # listen_address, listen_port = arg_parser() # распарсили аргументы запуска
    # database = ServerStorage() # создали объект подключения к DB
    server = MsgServer(listen_address, listen_port, database)  # объект сервера Worker-a
    server.daemon = True
    server.start()  # запустили поток воркера
    # console(server)

    # Создаём графическое окружение для сервера:
    server_app = QApplication(sys.argv)
    main_window = MainWindow()

    # Инициализируем параметры в окна
    main_window.statusBar().showMessage('Server Working')
    main_window.active_clients_table.setModel(gui_create_model(database))
    main_window.active_clients_table.resizeColumnsToContents()
    main_window.active_clients_table.resizeRowsToContents()

    def list_update():
        """
        Метод обновляющий список подключённых, проверяет флаг подключения, и
        если надо обновляет список
        """
        global new_connection
        # new_connection = True
        if new_connection:
            main_window.active_clients_table.setModel(gui_create_model(database))
            main_window.active_clients_table.resizeColumnsToContents()
            main_window.active_clients_table.resizeRowsToContents()
            with conflag_lock:
                new_connection = False

    def show_statistics():
        """Метод создающий окно со статистикой клиентов"""
        global stat_window
        stat_window = HistoryWindow()
        stat_window.history_table.setModel(create_stat_model(database))
        stat_window.history_table.resizeColumnsToContents()
        stat_window.history_table.resizeRowsToContents()
        stat_window.show()


    def server_config():
        """Метод создающий окно с настройками сервера"""
        global config_window
        # Создаём окно и заносим в него текущие параметры
        config_window = ConfigWindow()
        config_window.db_path.insert(config['SETTINGS']['Database_path'])
        config_window.db_file.insert(config['SETTINGS']['Database_file'])
        config_window.port.insert(config['SETTINGS']['Default_port'])
        config_window.ip.insert(config['SETTINGS']['Listen_Address'])
        config_window.save_btn.clicked.connect(save_server_config)

    def save_server_config():
        """Метод сохранения настроек"""
        global config_window
        message = QMessageBox()
        config['SETTINGS']['Database_path'] = config_window.db_path.text()
        config['SETTINGS']['Database_file'] = config_window.db_file.text()
        try:
            port = int(config_window.port.text())
        except ValueError:
            message.warning(config_window, 'Ошибка', 'Порт должен быть числом')
        else:
            config['SETTINGS']['Listen_Address'] = config_window.ip.text()
            if 1023 < port < 65536:
                config['SETTINGS']['Default_port'] = str(port)
                print(port)
                with open('server.ini', 'w') as conf:
                    config.write(conf)
                    message.information(
                        config_window, 'OK', 'Настройки успешно сохранены!')
            else:
                message.warning(
                    config_window,
                    'Ошибка',
                    'Порт должен быть от 1024 до 65536')


    # Таймер, обновляющий список клиентов 1 раз в секунду
    timer = QTimer()
    timer.timeout.connect(list_update)
    timer.start(1000)

    # Связываем кнопки с процедурами
    main_window.refresh_button.triggered.connect(list_update)
    main_window.show_history_button.triggered.connect(show_statistics)
    main_window.config_btn.triggered.connect(server_config)
    main_window.register_btn.triggered.connect(server.reg_user)
    main_window.remove_btn.triggered.connect(server.rem_user)

    # Запускаем GUI

    server_app.exec_()


class MsgServer(threading.Thread, metaclass=ServerVerifier):
    """main server messaging class"""
    # дескриптор порта
    listen_port = IsPortValid()

    def __init__(self, listen_address, listen_port, database):
        self.clients = []
        self.messages = []
        self.names = dict()
        self.listen_address = listen_address
        self.listen_port = listen_port
        self.database = database
        if not 1023 < self.listen_port < 65535:
            LOGGER.critical(f'Невозможно запустить сервер на порту {self.listen_port}, порт занят или недопустим')
            sys.exit(1)
        super(MsgServer, self).__init__()

    def kill_server(self):  # todo вышибать  процесс сервера при занятии порта по эксепшену
        """Метод выключения сервера. Фильтрует процессы по порту и убивает"""
        # new_ping = subprocess.Popen(item, stdout=subprocess.PIPE)
        # for line in new_ping.stdout:
        #     result = chardet.detect(line)
        #     line = line.decode(result['encoding']).encode('utf-8')
        #     print(line.decode('utf-8'))

        that = ["netstat", "-aon", "|", "findstr", self.listen_port]
        # if platform.system() == 'Windows':
        str = subprocess.Popen(that, stdout=subprocess.PIPE)
        # else:
        #    str =  subprocess.Popen(that, stdout=subprocess.PIPE)
        result = chardet.detect(str)
        for line in str.stdout:
            print(line)
            line = line.decode(result['encoding']).encode('utf-8')
            print(line)

    def process_client_message(self, message, client):
        """ метод разбирающий клиентские сообщения. Принимает на вход словарь сообщения, проверяет их корректность,
         ничего не возвращает, отравляет ответ клиенту в виде словаря
        """
        global new_connection
        LOGGER.debug(f'Попытка разобрать клиентское сообщение: {message}')
        if ACTION in message and message[ACTION] == PRESENCE and TIME in message and USER in message:
            self.autorize_user(message, client)
            return

        # Если это запрос активных клиентов
        if ACTION in message and message[ACTION] == GETCLIENTS and TIME in message and USER in message:
            # запрашиваем список подключеных клиентов
            user_list = list(self.names.keys())
            # print(user_list)
            # если клиенты есть добавляем строку в ответ, если нет - возвращаем 204
            if user_list != '':
                RESPONSE_CLIENTS[RESPONSE] = 201
                RESPONSE_CLIENTS[LIST] = user_list
            else:
                RESPONSE_CLIENTS[RESPONSE] = 204
                LOGGER.debug(f'Нет клиентов подключеных к серверу')
            try:
                send_message(client, RESPONSE_CLIENTS)
                LOGGER.debug(f'Отправка списка клиентов подключеных к серверу {RESPONSE_CLIENTS}')
                return
            except Exception as err:
                print(1, err)

        # если пришел EXIT:
        elif ACTION in message and message[ACTION] == EXIT and ACCOUNT_NAME in message:
            with conflag_lock:
                self.database.user_logout(message[ACCOUNT_NAME])
                print(f'Отключен клиент {message[SENDER]}')
                this_one = self.names[message[ACCOUNT_NAME]]
                self.clients.remove(this_one)
                this_one.close()
                del this_one
                # with conflag_lock:
                new_connection = True
                LOGGER.info(f'Клиент {message[SENDER]} корректно отключен от сервера')
            return

        # Если это запрос контакт-листа
        elif ACTION in message and message[ACTION] == GETCONTACTS and USER in message and \
                self.names[message[USER]] == client:
            response = RESPONSE_202
            response[LIST] = self.database.get_user_contacts(message[USER])
            try:
                send_message(client, response)
                LOGGER.info(f'Клиенту {message[SENDER]} отправлен словарь контактов {response[LIST]}')
                return
            except Exception as err:
                print(1, err)

            # Если это добавление контакта

        # Если это добавление контакта
        elif ACTION in message and message[ACTION] == ADD_CONTACT and ACCOUNT_NAME in message and USER in message \
             and self.names[message[USER]] == client:
            self.database.add_contact(message[USER], message[ACCOUNT_NAME])
            send_message(client, RESPONSE_200)

        # Если это удаление контакта
        elif ACTION in message and message[ACTION] == REMOVE_CONTACT and ACCOUNT_NAME in message and USER in message \
                and self.names[message[USER]] == client:
            self.database.del_contact(message[USER], message[ACCOUNT_NAME])
            send_message(client, RESPONSE_200)

        # если пришло сообщение добавляем его в очередь
        elif ACTION in message and message[ACTION] == MESSAGE and TIME in message and MESSAGE_TEXT in message[USER] \
                and DESTINATION in message and SENDER in message:
            # print(message)
            LOGGER.debug(f"От клиета {message[SENDER]} получено сообщение {message[USER][MESSAGE_TEXT]}")
            self.messages.append(message)
            self.database.process_message(
                message[SENDER], message[DESTINATION])
            # send_message(client, RESPONSE_200)
            return

        # Если это запрос публичного ключа пользователя
        elif ACTION in message and message[ACTION] == PUBLIC_KEY_REQUEST and ACCOUNT_NAME in message[USER]:
            response = RESPONSE_511
            response[DATA] = self.database.get_pubkey(message[USER][ACCOUNT_NAME])
            # может быть, что ключа ещё нет (пользователь никогда не логинился,
            # тогда шлём 400)
            if response[DATA]:
                try:
                    send_message(client, response)
                except OSError:
                    self.remove_client(client)
            else:
                response = RESPONSE_400
                response[ERROR] = 'Нет публичного ключа для данного пользователя'
                try:
                    send_message(client, response)
                except OSError:
                    self.remove_client(client)

        # если ничего не подошло:
        else:
            LOGGER.debug(f"{message} Некорректный запрос, вернуть 400")
            response = RESPONSE_400
            response[ERROR] = 'Некорректный запрос'
            send_message(client, response)
            return

    def process_message(self, message, to_send_data_list):
        """
        метод адресной отправки сообщения определённому клиенту, принимает на вход сообщение,
        и слушающие сокеты. Ничего не возвращает.
        Вызывает отправку сообщения нужному клиенту
        """
        if message[DESTINATION] in self.names.keys() and self.names[message[DESTINATION]] in to_send_data_list:
            # print(self.names[message[DESTINATION]])
            send_message(self.names[message[DESTINATION]], message)
            LOGGER.info(f'Отправлено сообщение пользователю {message[DESTINATION]} '
                        f'от пользователя {message[SENDER]}.')

        elif message[DESTINATION] in self.names and self.names[message[DESTINATION]] not in to_send_data_list:
            raise ConnectionError
        else:

            LOGGER.error(
                f'Пользователь {message[DESTINATION]} не зарегистрирован на сервере, '
                f'отправка сообщения невозможна.')
            LOGGER.debug(
                f'на сервере остались {self.names.keys()} ')

    def autorize_user(self, message, client):
        """Метод реализующий авторизцию пользователей."""
        # Если имя пользователя уже занято-то возвращаем 400
        LOGGER.debug(f'Start auth process for {message[USER]}')
        if message[USER][ACCOUNT_NAME] in self.names.keys():
            response = RESPONSE_400  # todo добавить отключение неактивных reverse_ping
            response[ERROR] = 'Имя пользователя уже занято.'
            try:
                LOGGER.debug(f'Имя пользователя уже занято {response}')
                send_message(client, response)
            except OSError:
                LOGGER.debug('OS Error')
                pass
            self.clients.remove(client)
            client.close()
        # Проверяем что пользователь зарегистрирован на сервере.
        elif not self.database.check_user(message[USER][ACCOUNT_NAME]):
            response = RESPONSE_400
            response[ERROR] = 'Пользователь не зарегистрирован.'
            try:
                LOGGER.debug(f'Unknown username, sending {response}')
                send_message(client, response)
            except OSError:
                pass
            self.clients.remove(client)
            client.close()
        else:
            LOGGER.debug('Пользователь получен, начинаем аутентификацию.')
            # если все ок - отвечаем 511 и запускаем аутентицикацию
            auth_message = RESPONSE_511
            # рандомная строка в hex представлении соль для проверки ключа
            random_str = binascii.hexlify(os.urandom(64))
            # В словарь байты нельзя, декодируем (json.dumps -> TypeError)
            auth_message[DATA] = random_str.decode('ascii')
            # взяли хеш пароля из базы и хешируем с ранее полученой рандомной строкой, сохраняем серверную версию ключа
            hash = hmac.new(self.database.get_hash(message[USER][ACCOUNT_NAME]), random_str, 'MD5')
            # для красоты и удобства получаем дайджест
            digest = hash.digest()
            LOGGER.debug(f'соль для клиента = {auth_message}')
            try:
                # отсылаем клиенту соль
                send_message(client, auth_message)
                ans = get_message(client)
            except OSError as err:
                LOGGER.debug('Ошибка проверки подлинности, data:', exc_info=err)
                client.close()
                return
            # полученый ответ клиента преобразуем из ascii  в дайджест
            client_digest = binascii.a2b_base64(ans[DATA])
            # Сверяем то что прислал клиент с тем что вычислил сервер.
            # compare_digest используем чтобы предотвратить атаку по веремени
            if RESPONSE in ans and ans[RESPONSE] == 511 and hmac.compare_digest(
                    digest, client_digest):
                LOGGER.debug(f"От клиента пришло  {ans}")
                # если все в порядке до добавляем клиента в список активных
                self.names[message[USER][ACCOUNT_NAME]] = client
                client_ip, client_port = client.getpeername()
                # и шлем ему 200-OK
                try:
                    send_message(client, RESPONSE_200)
                except OSError:
                    # если не получилось отправить 200, то удаляем, значит не судьба
                    self.remove_client(message[USER][ACCOUNT_NAME])
                # добавляем пользователя в список активных и если у него изменился открытый ключ
                # сохраняем новый
                new_pub_key = self.database.get_pubkey(message[USER][ACCOUNT_NAME])
                if message[USER][PUBLIC_KEY] != new_pub_key:
                    self.database.user_login(
                        message[USER][ACCOUNT_NAME],
                        client_ip,
                        client_port,
                        message[USER][PUBLIC_KEY])
                message = {}
                return
            else:
                response = RESPONSE_400
                response[ERROR] = 'Неверный пароль.'
                try:
                    send_message(client, response)
                except OSError:
                    pass
                self.clients.remove(client)
                client.close()

    def service_update_lists(self):
        """ Метод реализующий отправки сервисного сообщения 205 клиентам. """
        for client in self.names:
            try:
                send_message(self.names[client], RESPONSE_205)
            except OSError:
                self.remove_client(self.names[client])

    def run(self):
        """Метод запуска поотка обработки сообщения"""
        global new_connection
        LOGGER.info('Попытка запуска сервера')
        try:
            transport = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # transport = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # проверка метакласса
            transport.bind((self.listen_address, self.listen_port))
            transport.listen(MAX_CONNECTIONS)
            transport.settimeout(0.1)
            # transport.connect('192.168.22.1', '80') # проверка метакласса

        except OSError as err:
            LOGGER.error(
                f'Адрес {self.listen_address} и порт {self.listen_port} не  могут быть использованы для запуска,'
                f' потому что уже используются другой программой', err)
            sys.exit(1)
        else:
            print(
                f'Запущен сервер прослушивающий на {self.listen_address if self.listen_address else "любом"} '
                f'ip-адресе и {self.listen_port} порту')
            LOGGER.info(
                f'Запущен сервер прослушивающий на {self.listen_address if self.listen_address else "любом"} ip-адресе'
                f' и {self.listen_port} порту')
        # self.kill_server()

        while True:
            # Принимаем подключения
            try:
                client, client_address = transport.accept()
            except OSError:
                pass
            else:
                LOGGER.info(f'Установлено соединение с адресом {client_address}')
                self.clients.append(client)
            incoming_data_list = []
            to_send_data_list = []
            errors_list = []

            # проверяем есть ли новые клиенты или данные от уже подключеных
            try:
                if self.clients:
                    incoming_data_list, to_send_data_list, errors_list = select.select(self.clients, self.clients, [],
                                                                                       0)
            except OSError:
                pass

            # парсим сообщения клиентов и если есть сообщения кладем их в словарь сообщений,
            # если ошибка - исключаем клиента
            if incoming_data_list:
                for sended_from_client in incoming_data_list:
                    # print(get_message(sended_from_client))
                    try:
                        self.process_client_message(get_message(sended_from_client), sended_from_client)
                    except Exception as err:
                        print(err)

                        LOGGER.info(f'{sended_from_client.getpeername()} отключился от сервера')
                        new_connection = True
                        self.clients.remove(sended_from_client)

            for one_message in self.messages:
                try:
                    LOGGER.debug(f'Обработка сообщения {one_message}')
                    self.process_message(one_message, to_send_data_list)
                except Exception:
                    LOGGER.info(f'Соединение с {one_message[DESTINATION]} разорвано')
                    try:
                        self.clients.remove(self.names[one_message[DESTINATION]])
                        del self.names[one_message[DESTINATION]]
                    except Exception:
                        pass  # todo добавить отбойник что сообщение не доставлено, или сохранение в очередь
            self.messages.clear()

    def remove_client(self, client):
        """Метод удаления клиента"""
        LOGGER.info(f'Клиент {client.getpeername()} отключился от сервера.')
        for name in self.names:
            if self.names[name] == client:
                self.database.user_logout(name)
                del self.names[name]
                break
        self.clients.remove(client)
        client.close()
        return

    def reg_user(self):
        """Метод создающий окно регистрации пользователя."""
        global reg_window
        reg_window = RegisterUser(self.database, self)
        reg_window.show()

    def rem_user(self):
        """Метод создающий окно удаления пользователя."""
        global rem_window
        rem_window = DelUserDialog(self.database, self)
        rem_window.show()


if __name__ == '__main__':
    main()



"""Константы"""

# Порт по умолчанию для сетевого взаимодействия
DEFAULT_PORT = 7777
# IP адрес по умолчанию для подключения клиента
DEFAULT_IP_ADDRESS = '127.0.0.1'
# Максимальная очередь подключений
MAX_CONNECTIONS = 5
# Максимальная длинна сообщения в байтах
MAX_PACKAGE_LENGTH = 4096
# Кодировка проекта
ENCODING = 'utf-8'
# Уровень логирования
# LOGGING_LEVEL = 'INFO'
LOGGING_LEVEL = 'DEBUG'


# Прококол JIM основные ключи:
ACTION = 'action'
TIME = 'time'
USER = 'user'
ACCOUNT_NAME = 'account_name'
SENDER = 'sender'
DESTINATION = 'destination'
DATA = 'bin'
PUBLIC_KEY = 'pubkey'


# Прочие ключи, используемые в протоколе
PRESENCE = 'presence'
RESPONSE = 'response'
ERROR = 'error'
MESSAGE = 'message'
GETCLIENTS = 'getclients'
GETCONTACTS = 'get_contacts'
ADD_CONTACT = 'add_contact'
REMOVE_CONTACT = 'remove_contact'
USERS_REQUEST = 'get_users'
MESSAGE_TEXT = 'message_text'
MESSAGE_KEY = 1
ACCOUNT_KEY = 0
EXIT = 'exit'
STATUS = 'status_message'
LIST = 'user_list'
MYCOLOR = '0900FF'
NOTMYCOLOR = '000000'
PUBLIC_KEY_REQUEST = 'get_public_key'

# Словари - ответы:
# 200
RESPONSE_200 = {RESPONSE: 200}
# 400
RESPONSE_400 = {
    RESPONSE: 400,
    ERROR: None
}
RESPONSE_204 = {RESPONSE: 204}
RESPONSE_202 = {RESPONSE: 202}
RESPONSE_205 = {RESPONSE: 205}
RESPONSE_CLIENTS = {
    RESPONSE: 201
}
RESPONSE_511 = {
    RESPONSE: 511,
    DATA: None
}



from setuptools import find_packages, setup

setup(name="bervr messenger server",
      version="2.1.8",
      description="A server part for messenger",
      author="Aleksandr Filippov",
      author_email="bervrr@gmail.com",
      packages=find_packages(),
      # package_data={
      #       'server': ['*'],
      #       './server/server': ['*']
      # },
      include_package_data = True,
      install_requires=['PyQt5', 'SQLAlchemy', 'pycryptodome', 'pycryptodomex', 'chardet'],
      scripts=['./server/server_run.py']
      )



"""Утилиты"""
import argparse
import json

from common.variables import MAX_PACKAGE_LENGTH, ENCODING, DEFAULT_PORT, DEFAULT_IP_ADDRESS
from common.decor import func_log, Log


@Log()
def get_message(client):
    """Функция декодирующая сообщение из байтов в JSON"""
    encoded_response = client.recv(MAX_PACKAGE_LENGTH)
    if isinstance(encoded_response, bytes):
        json_response = encoded_response.decode(ENCODING)
        response = json.loads(json_response)
        if isinstance(response, dict):
            return response
        raise ValueError
    raise ValueError


@func_log
def send_message(sock, message):
    """Функция кодирующаяя сообщение для отправки из JSON в байты"""
    js_message = json.dumps(message)
    encoded_message = js_message.encode(ENCODING)
    sock.send(encoded_message)


def create_arg_parser(default_port, default_address):
    """Функция парсер аргументов командной строки"""
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument('-p', default=default_port, type=int, nargs='?')
        parser.add_argument('-a', default=default_address, nargs='?')
        parser.add_argument('-m', default='listen', nargs='?')
        parser.add_argument('-n', default='', nargs='?')
        parser.add_argument('-w', default='', nargs='?')
        return parser
        if server_port < 1024 or server_port > 65535:
            raise ValueError
    except ValueError:
        return 'В качестве порта может быть указано только число в диапазоне от 1024 до 65535.'




import dis
import socket


class ServerVerifier(type):
    """
    Метакласс ServerVerifier, выполняет базовую проверку класса «Сервер»:
    отсутствие вызовов connect для сокетов и использование сокетов для работы по TCP
    """

    def __new__(cls, name, bases, dict):
        new_class = super(ServerVerifier, cls).__new__(cls, name, bases, dict)
        for func in new_class.__dict__:
            try:
                ret = dis.get_instructions(new_class.__dict__[func])
            except TypeError:
                pass
            except SyntaxError:
                pass
            else:
                for i in ret:
                    if i.argval == 'connect':
                        raise Exception(f"Method connect can't be used in this module: {func}")
                    elif i.argval == 'SOCK_DGRAM':
                        raise Exception(f"UDP connection not supported in this version. Method: {func}")
        return new_class


class ClientVerifier(type):
    """
    Метакласс ClientVerifier, выполняет базовую проверку класса «Клиент»
    отсутствие вызовов accept и listen для сокетов;
    использование сокетов для работы по TCP;
    отсутствие создания сокетов на уровне классов, то есть отсутствие конструкций такого вида:
    class Client: s = socket(). Для проверок используется модуль dis:
    """

    def __new__(cls, name, bases, dict):
        client_class = super(ClientVerifier, cls).__new__(cls, name, bases, dict)
        for func, value in client_class.__dict__.items():
            if type(value) is type(socket.socket()):
                raise Exception(f"Creation socket in this level is forbidden: {func}")
            try:
                ret = dis.get_instructions(client_class.__dict__[func])
            except TypeError:
                pass
            except SyntaxError:
                pass
            else:
                for i in ret:
                    if i.argval == 'accept' or i.argval == 'listen':
                        raise Exception(f"Accept and listen method's can't be used in this module: {func}")
                    elif i.argval == 'SOCK_DGRAM':
                        raise Exception(f"UDP connection not supported in this version. Method: {func}")
        return client_class



"""Параметры логирования сервера"""

import logging
import logging.handlers
import os
import sys
# дабавим путь для импорта переменной
sys.path.append(os.path.join(os.getcwd(), '../../../../messenger/server'))
from common.variables import LOGGING_LEVEL

# задаем путь куда будем писать
PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir)
PATH = os.path.join(PATH, 'server.log')

# создали логгер
SERVER_LOGGER = logging.getLogger('server')

#
# LOGGER.setLevel(logging.INFO)

# уровень логгирования для данного хендлера
FILE_ROTATE_HANDLER = logging.handlers.TimedRotatingFileHandler(PATH, encoding='utf-8', interval=1, when='D')
# FILE_ROTATE_HANDLER.setLevel(logging.DEBUG)

# формат логгирования
FORMATTER = logging.Formatter("%(asctime)-25s %(levelname)-9s %(filename)-10s %(message)s")

# привязали форматирование к хендлеру
FILE_ROTATE_HANDLER.setFormatter(FORMATTER)

# привязали хендлер к логгеру
SERVER_LOGGER.addHandler(FILE_ROTATE_HANDLER)

# facility
SERVER_LOGGER.setLevel(LOGGING_LEVEL)

# еще один хедлер для вывода ошибок в консоль:
# создали хендлер
STREAM_HANDLER = logging.StreamHandler(sys.stdout)

# уровень логгирования для данного хендлера
STREAM_HANDLER.setLevel(logging.CRITICAL)

# привязали формалирование к хендлеру
STREAM_HANDLER.setFormatter(FORMATTER)

# привязали хендлер к логгеру
SERVER_LOGGER.addHandler(STREAM_HANDLER)

if __name__ == '__main__':
    SERVER_LOGGER.debug('test debug message')
    SERVER_LOGGER.info('test info message')
    SERVER_LOGGER.warning('test warning message')
    SERVER_LOGGER.critical('test critical message')
    SERVER_LOGGER.error('test error message')
